query ValidateBulkUploadStatus($input: ID!) {
    validateBulkUploadStatus(receiptId: $input) {
        status
        error
        result {
            issues {
                path
                errorMessage
            }
        }
    }
}