query GetRuleBody($input: ID!) {
    rulePythonBody(input: $input) {
        pythonBody
        tests {
            expectedResult
            name
            resource
            mocks {
                objectName
                returnValue
            }
        }
    }
}