mutation TestCorrelationRule($input: TestCorrelationRuleYAMLInput!) {
    testCorrelationRuleYAML(input: $input) {
        results {
            error
            name
            passed
        }
    }
}
